unit ProcMath;

interface

uses Windows,LibV, tgen;

const
  SIZE_POWER = 13;
  INBLK_SIZE = 8192;
  PAD_LEFT	= 1024;
  PAD_RIGHT = 1024;
  DATA_LEN = (INBLK_SIZE-PAD_LEFT-PAD_RIGHT);

{$DEFINE DEBF}

function BeginProcessing(Frequencies,Amplitudes,
  Channel1Window,Channel2Window : vector_p):vboolean;

procedure EndProcessing;

procedure IdleProc(Processing:vboolean;var Done:Boolean);

implementation

var
  waveout : vhandle_t = LIBV_INVALID_HANDLE;

  Freq, Amp, WinLeft, WinRight : vector_p;


  Samples : array[0..INBLK_SIZE-1] of vec_t;

  InReal: array[0..INBLK_SIZE-1] of vec_t;
  Dummy : array[0..INBLK_SIZE-1] of vec_t;

  FreqReal : array[0..INBLK_SIZE-1] of vec_t;
  FreqImag : array[0..INBLK_SIZE-1] of vec_t;

  FreqReal1 : array[0..INBLK_SIZE-1] of vec_t;
  FreqImag1 : array[0..INBLK_SIZE-1] of vec_t;

  FreqReal2 : array[0..INBLK_SIZE-1] of vec_t;
  FreqImag2 : array[0..INBLK_SIZE-1] of vec_t;

  Samples1 : array[0..INBLK_SIZE-1] of vec_t;
  Samples2 : array[0..INBLK_SIZE-1] of vec_t;

  OutSamples : array[0..2*INBLK_SIZE-1] of s16_t;

// This routine stores float data from 'Samples1' and 'Samples2'
// in packed 16-bit PCM format

procedure Collect2Channels;assembler;
const
  upper_limit : vec_t = MAXSHORT;
asm
  push ebx
  push esi

  mov esi, DATA_LEN
  mov eax, PAD_LEFT
  shl eax, 2
  mov ebx, offset Samples1
  mov ecx, offset Samples2
  add ebx, eax
  add ecx, eax
  dec esi
  mov edx, offset OutSamples

@@next_pair:
  fld dword ptr[ebx + esi * 4]
  fld upper_limit
  fmulp
  fcom upper_limit
  fstsw ax
  sahf
  jbe @@_1_ok
  ffree st(0)
  mov word ptr[edx + esi * 4], MAXSHORT
  jmp @@_load_2
@@_1_ok:
  fistp word ptr[edx + esi * 4]
@@_load_2:
  fld dword ptr[ecx + esi * 4]
  fld upper_limit
  fmulp
  fcom upper_limit
  fstsw ax
  sahf
  jbe @@_2_ok
  ffree st(0)
  mov word ptr[edx + esi * 4 + 2], MAXSHORT
  jmp @@_check_end
@@_2_ok:
  fistp word ptr[edx + esi * 4 + 2]
@@_check_end:
  dec esi
  jge @@next_pair

  pop esi
  pop ebx
end;

procedure FFTForward;
begin
  memZero(@Dummy,INBLK_SIZE * sizeof(vec_t));
  fftCalcNip(@FreqReal,@FreqImag,@InReal,@Dummy,SIZE_POWER,1);
end;

procedure ApplyWindows;
var
  wlen,i : integer;
begin
  wlen := INBLK_SIZE-1;
  for i:=0 to wlen do
  begin
    FreqReal1[i] := FreqReal[i] * WinLeft^.elements^[i];
		FreqImag1[i] := FreqImag[i] * WinLeft^.elements^[i];
    FreqReal2[i] := FreqReal[i] * WinRight^.elements^[i];
		FreqImag2[i] := FreqImag[i] * WinRight^.elements^[i];
  end;
end;

procedure FFTInverse;
begin
  fftCalcNip(@Samples1,@Dummy,@FreqReal1,@FreqImag1,SIZE_POWER,-1);
  fftCalcNip(@Samples2,@Dummy,@FreqReal2,@FreqImag2,SIZE_POWER,-1);
end;

{$IFDEF DEBF}
var
  wfile : vhandle_t;
{$ENDIF}

procedure WriteStereoOutput;
begin
  LibV.fileWrite(waveout,@OutSamples,DATA_LEN * 4);
{$IFDEF DEBF}
  LibV.fileWrite(wfile,@OutSamples,DATA_LEN * 4);
{$ENDIF}
end;

var
  u : vec_t;

procedure FFTFixUp;
var
  i : integer;
  delta, du : vec_t;
begin
  delta := u - Samples1[PAD_LEFT];
  if {v_abs(delta) < 0.2} true then
  begin
    du := delta / DATA_LEN;
    delta := du;
    i := PAD_LEFT+DATA_LEN-2;
    repeat
      Samples1[i] := Samples1[i]+delta;
      delta := delta + du;
      i := i - 1;
    until i < PAD_LEFT;
  end;
  u := Samples1[PAD_LEFT+DATA_LEN-1];
end;

procedure FFTPad;
var
  i, start, limit : integer;
begin
  limit := PAD_LEFT + PAD_RIGHT;
  { First block: PAD_LEFT + PAD_RIGHT copies of last samples }
  i := 0;
  start := DATA_LEN;
  while i <> limit do
  begin
    InReal[i] := InReal[start + i];
    i := i + 1;
  end;
  { Rest of data: new samples }
  repeat
    InReal[i] := Samples[i];
    i := i + 1;
  until i = INBLK_SIZE;
end;

procedure ProcessNextBlock;
begin
  toneGenerate(@Samples[PAD_LEFT+PAD_RIGHT],DATA_LEN);
  FFTPad;
  FFTForward;
  ApplyWindows;
  FFTInverse;
  FFTFixUp;
  Collect2Channels;
  WriteStereoOutput;
end;

procedure IdleProc(Processing:vboolean;var Done:Boolean);
const
  max_qlen = 44100 * 5;
var
  qlen : u32_t;
begin
  if not Processing then
  begin
    Done := True;
    Exit;
  end;
  qlen := 0;
  if fileControl(waveout,pctl_waveout_get_pending_samples,qlen) then
    if qlen < max_qlen then
      ProcessNextBlock;
  Done := False;
end;

function BeginProcessing(Frequencies,Amplitudes,
  Channel1Window,Channel2Window : vector_p):vboolean;
begin
  Result := False;
  if not toneInit(Frequencies,Amplitudes) then
    Exit;
  if waveout = LIBV_INVALID_HANDLE then
  begin
    Windows.MessageBox(NULL,
      'Multipass filter','Unable to open waveform output',MB_ICONHAND);
    Exit;
  end;
  Freq := Frequencies;
  Amp := Amplitudes;
  WinLeft := Channel1Window;
  WinRight := Channel2Window;
  Result := True;
end;

procedure EndProcessing;
begin
  wavePlaybackReset(waveout);
end;

initialization
  LibV.logStartConsoleLog();
  LibV.logUpperLevel(LOG_TRACE);
  waveout := LibV.wavePlaybackOpen(2,44100,16);
{$IFDEF DEBF}
  wfile := LibV.wavOpenWrite('//d/out.wav',2,44100,16);
{$ENDIF}
finalization
  LibV.fileClose(waveout);
{$IFDEF DEBF}
  LibV.fileClose(wfile);
{$ENDIF}
end.

